#!/usr/bin/runhugs +l

> module Main(main) where

> import CGI

> import Maybe
> import Random
> import IO

> import Syntax
> import HParse
> import Matching
> import Calculations
> import Laws
> import Types
> import Constants
> import Rewrite
> import PPrint

> import State
> import Exception


> tmppath = "/tmp/"
> examplespath = "/home/ganesh/mag/examples/"
> examples = [("fast reverse","reverse"),
>             ("postorder traversal","postorder"),
>             ("minimum depth","mindepth"),
>             ("alpha beta pruning","alphabeta"),
>             ("fast abstracts","abstracts"),
>             ("fibonacci sequence","fib"),
>             ("steep sequences","steep"),
>             ("longest path length","llp")]

> main :: IO ()
> main = wrapper (\env -> do body <- handleerrs $ cgimag env
>                            return $ Content {mime = page "MAG" [] body})
>   where handleerrs m 
>           = catchEx (trapIOEx m 
>                               (\ioerr -> 
>                                   return [p $ quote $ show ioerr]))
>                     return

> cgimag :: [(Name,Value)] -> ExT [HTML] IO [HTML]

> cgimag params 
>    = case maybe "enter" id (lookup "mode" params) of
>         "enter" -> enterpage (lookup "preload" params)
>         "load" -> do source <- transformExUnit $ getparam "source" params
>                      loaddata source
>         "transform" -> 
>            do transform <- transformExUnit $ getparam "transform" params
>               name <- transformExUnit $ getparam "name" params
>               transformapply transform name

> getparam :: Name -> [(Name,Value)] -> Ex [HTML] Value
> getparam name env 
>    = case lookup name env of
>       Just value -> return value
>       Nothing -> raiseEx [p [prose ("Must supply "++name++" parameter")]]

> enterpage preload
>   = do scriptname <- liftEx $ getScriptName
>        source <- 
>           case preload of
>             Nothing -> return ""
>             Just name -> liftEx $
>               do handle <- openFile (examplespath ++ name ++ ".p") ReadMode
>                  hGetContents handle
>        return $ prose "examples: "
>               : sepwith (prose ",") (map (preloadlink scriptname) examples)
>              ++ [gui scriptname [p [textarea "source" 30 80 "yes" source],
>                                  hidden "mode" "load",
>                                  p [submit "load" "Parse and type check"]]]
>    where preloadlink scriptname (text,file) 
>           = href (geturl scriptname [("preload",file)]) [prose text]

> sepwith s [] = []
> sepwith s [x] = [x]
> sepwith s (x:xs) = x:s:sepwith s xs

> loaddata :: String -> ExT [HTML] IO [HTML]
> loaddata source
>  = do sfi <- transformExUnit $ checkdata source
>       let sfi' = sfi { sfitext = source }
>       storeandpresentdata [p [prose "OK, program loaded"]] sfi'

> storeandpresentdata :: [HTML] -> SourceFileInfo -> ExT [HTML] IO [HTML]
> storeandpresentdata heading sfi  
>          = do tmpname <- sequence $ take 8 $ repeat $ liftEx
>                                   $ getStdRandom (randomR ('a','z'))
>               tmpfile <- liftEx $ openFile (tmppath++tmpname++".p") WriteMode
>               let text = sfitext sfi
>               liftEx $ hPutStr tmpfile text
>               scriptname <- liftEx $ getScriptName
>               return $ heading
>                     ++ quote text
>                     ++ [p [prose "Apply transform:"]]
>                     ++ map (makelink scriptname tmpname) (sfitransforms sfi)
>    where makelink scriptname tmpname trafo 
>           = p [href (geturl scriptname [("mode","transform"),
>                                         ("transform",trafo),
>                                         ("name",tmpname)])
>                     [prose trafo]]

> loadfile :: String -> ExT [HTML] IO SourceFileInfo
> loadfile file = do handle <- liftEx $ openFile (fullpath file) ReadMode
>                    source <- liftEx $ hGetContents handle
>                    transformExUnit $ checkdata source
>           where fullpath name = tmppath++name++".p"

> geturl :: URL -> [(Name,Value)] -> URL
> geturl url params = url ++ querystring params
>   where querystring [] = ""
>         querystring params = '?':ampjoin (map equalsify params)
>         ampjoin [str] = str
>         ampjoin (str:strs) = str ++ "&" ++ ampjoin strs
>         equalsify (name,value) = name ++ "=" ++ value

> quote :: String -> [HTML]
> quote str = [pre str]
> {-
> quote s = let (lines,rem) = quote' s in prose rem:lines
>  where quote' ('\n':s) 
>           = let (lines,rem) = quote' s 
>             in (unclosedElement "BR" []:prose rem:lines,[])
>        quote' (c:s) 
>           = let (lines,rem) = quote' s
>             in (lines,c:rem)
>        quote' [] = ([],"")
> -}

> checkdata :: String -> Ex [HTML] SourceFileInfo
> checkdata source
>   = if not $ null parseerrs
>     then raiseEx $ quote parseerrs
>     else checkdata' parsed
>       where (parsed,parseerrs) = parse' source


> checkdata' :: SourceFile -> Ex [HTML] SourceFileInfo
> checkdata' parsed 
>   = if not $ null tcerrs 
>     then raiseEx $ concat $ map quote tcerrs
>     else return tced
>       where (tced,tcerrs) = typecheck' parsed

> rewriter :: Theor -> (String,Type) -> St [HTML] [Def]
> rewriter th (name,typ) 
>     = do let exp = Con (Constr name) typ
>          let calc = calculate fastmatcher th (exp,[])
>          current <- readS
>          writeS (current ++ quote (prettyCalc 80 0 calc))
>          return [Fun (Var name typ) (unfreeze $ conclusion calc)]
>  where unfreeze (Lam p b t) = Lam p (unfreeze b) t
>        unfreeze (Ap f a t) = Ap (unfreeze f) (unfreeze a) t
>        unfreeze (Var x t) = Var x t
>        unfreeze (Con (Constr x) t) = Var x t
>        unfreeze (Con c t) = Con c t

> transformapply :: String -> String -> ExT [HTML] IO [HTML]
> transformapply trafo file
>    = do sfi <- loadfile file
>         ta <- fromJustEx [p [prose (trafo++" not found")]]
>                   $ sfitrafomap sfi trafo
>         let (newsourcefile,derivs) = eval [] $
>              do newsourcefile <- applytrafo ta rewriter
>                 derivs <- readS
>                 return (newsourcefile,derivs)
>         nsfi <- transformExUnit $ checkdata' newsourcefile
>         let nsfi' = nsfi { sfitext = prettySourceFile (sfifile nsfi) }
>         storeandpresentdata derivs nsfi'

> fromJustEx err (Just v) = return v
> fromJustEx err (Nothing) = raiseEx err
